from collections import defaultdict
from typing import BinaryIO


def read_batch_from_end(byte_stream: BinaryIO,
                        *,
                        size: int,
                        end_position: int) -> bytes:
    """
    Reads batch from the end of given byte stream.

    >>> import io
    >>> byte_stream = io.BytesIO(b'Hello\\nWorld!')
    >>> read_batch_from_end(byte_stream, size=4, end_position=5)
    b'ello'
    """
    if end_position > size:
        offset = end_position - size
    else:
        offset = 0
        size = end_position
    byte_stream.seek(offset)
    return byte_stream.read(size)


# in bytes
code_units_sizes = defaultdict(lambda: 1,
                               {'utf_16': 2,
                                'utf_16_be': 2,
                                'utf_16_le': 2,
                                'utf_32': 4,
                                'utf_32_be': 4,
                                'utf_32_le': 4})



import codecs
import functools
import io
import itertools
import os
import sys
from collections import abc
from operator import (methodcaller,
                      sub)
from typing import (Any,
                    BinaryIO,
                    Iterable,
                    List,
                    Optional,
                    Sequence,
                    TextIO,
                    overload)

from .arithmetical import ceil_division
from .hints import (Domain,
                    Range)
from .textual import (code_units_sizes,
                      read_batch_from_end)


@overload
def reverse(object_: Sequence[Domain]) -> Sequence[Domain]:
    pass


if sys.version_info >= (3, 6):
    from typing import Reversible


    @overload
    def reverse(object_: Reversible[Domain]) -> Iterable[Domain]:
        pass


@overload
def reverse(object_: TextIO,
            *,
            batch_size: Optional[int] = ...,
            lines_separator: Optional[str] = ...,
            keep_lines_separator: bool = ...) -> Iterable[str]:
    pass


@overload
def reverse(object_: BinaryIO,
            *,
            batch_size: Optional[int] = ...,
            lines_separator: Optional[bytes] = ...,
            keep_lines_separator: bool = ...,
            code_unit_size: int = ...) -> Iterable[bytes]:
    pass


@overload
def reverse(object_: Domain) -> Range:
    pass


@functools.singledispatch
def reverse(object_: Domain, **_: Any) -> Range:
    """
    Returns reversed object.

    >>> list(reverse(range(10)))
    [9, 8, 7, 6, 5, 4, 3, 2, 1, 0]
    >>> import io
    >>> list(reverse(io.BytesIO(b'Hello\\nWorld!')))
    [b'World!', b'Hello\\n']
    """
    raise TypeError('Unsupported object type: {type}.'
                    .format(type=type(object_)))


@reverse.register(abc.Sequence)
def _(object_: Sequence[Domain]) -> Sequence[Domain]:
    """
    Returns reversed sequence.
    """
    return object_[::-1]


if sys.version_info >= (3, 6):
    from typing import Reversible


    @reverse.register(abc.Reversible)
    def _(object_: Reversible[Domain]) -> Iterable[Domain]:
        """
        Returns reversed reversible iterable.
        """
        yield from reversed(object_)


@reverse.register(io.TextIOWrapper)
def _(object_: TextIO,
      *,
      batch_size: Optional[int] = None,
      lines_separator: Optional[str] = None,
      keep_lines_separator: bool = True) -> Iterable[str]:
    """
    Returns reversed file object.
    """
    encoding = object_.encoding
    if lines_separator is not None:
        lines_separator = lines_separator.encode(encoding)
    code_unit_size = code_units_sizes[encoding]
    if batch_size is not None:
        batch_size = ceil_division(batch_size, code_unit_size) * code_unit_size
    yield from map(functools.partial(codecs.decode,
                                     encoding=encoding),
                   reverse(object_.buffer,
                           batch_size=batch_size,
                           lines_separator=lines_separator,
                           keep_lines_separator=keep_lines_separator,
                           code_unit_size=code_unit_size))


@reverse.register(io.BufferedReader)
@reverse.register(io.BytesIO)
def _(object_: BinaryIO,
      *,
      batch_size: Optional[int] = None,
      lines_separator: Optional[bytes] = None,
      keep_lines_separator: bool = True,
      code_unit_size: int = 1) -> Iterable[bytes]:
    """
    Returns reversed byte stream.
    """
    if lines_separator is None:
        lines_separator = (b'\r', b'\n', b'\r\n')
        lines_splitter = methodcaller(bytes.splitlines.__name__,
                                      keep_lines_separator)
    else:
        def lines_splitter(byte_sequence: bytes) -> List[bytes]:
            result = []
            part = bytearray()
            offset = 0
            add_part = result.append
            while offset < len(byte_sequence):
                if (byte_sequence[offset:offset + len(lines_separator)]
                        != lines_separator):
                    part += byte_sequence[offset:offset + code_unit_size]
                    offset += code_unit_size
                else:
                    add_part(part + keep_lines_separator * lines_separator)
                    part.clear()
                    offset += len(lines_separator)
            add_part(part)
            return result
    stream_size = object_.seek(0, os.SEEK_END)
    if batch_size is None:
        batch_size = stream_size or 1
    batches_count = ceil_division(stream_size, batch_size)
    remaining_bytes_indicator = itertools.islice(
            itertools.accumulate(itertools.chain([stream_size],
                                                 itertools.repeat(batch_size)),
                                 sub),
            batches_count)
    try:
        remaining_bytes_count = next(remaining_bytes_indicator)
    except StopIteration:
        return

    def read_batch(position: int) -> bytes:
        result = read_batch_from_end(object_,
                                     size=batch_size,
                                     end_position=position)
        while result.startswith(lines_separator):
            try:
                position = next(remaining_bytes_indicator)
            except StopIteration:
                break
            result = (read_batch_from_end(object_,
                                          size=batch_size,
                                          end_position=position)
                      + result)
        return result

    batch = read_batch(remaining_bytes_count)
    segment, *lines = lines_splitter(batch)
    yield from reverse(lines)
    for remaining_bytes_count in remaining_bytes_indicator:
        batch = read_batch(remaining_bytes_count)
        lines = lines_splitter(batch)
        if batch.endswith(lines_separator):
            yield segment
        else:
            lines[-1] += segment
        segment, *lines = lines
        yield from reverse(lines)
    yield segment



import functools
import itertools
from typing import (Iterable,
                    Tuple)

from .functional import (combine,
                         compose)
from .hints import (Domain,
                    Map,
                    Operator,
                    Predicate)
from .replication import duplicate as _duplicate


def sifter(predicate: Predicate = None) -> Operator[Iterable[Domain]]:
    """
    Returns function that selects elements from iterable
    which satisfy given predicate.

    If predicate is not specified than true-like objects are selected.

    >>> to_true_like = sifter()
    >>> list(to_true_like(range(10)))
    [1, 2, 3, 4, 5, 6, 7, 8, 9]

    >>> def is_even(number: int) -> bool:
    ...     return number % 2 == 0
    >>> to_even = sifter(is_even)
    >>> list(to_even(range(10)))
    [0, 2, 4, 6, 8]
    """
    return functools.partial(filter, predicate)


def scavenger(predicate: Predicate = None) -> Operator[Iterable[Domain]]:
    """
    Returns function that selects elements from iterable
    which dissatisfy given predicate.

    If predicate is not specified than false-like objects are selected.

    >>> to_false_like = scavenger()
    >>> list(to_false_like(range(10)))
    [0]

    >>> def is_even(number: int) -> bool:
    ...     return number % 2 == 0
    >>> to_odd = scavenger(is_even)
    >>> list(to_odd(range(10)))
    [1, 3, 5, 7, 9]
    """
    return functools.partial(itertools.filterfalse, predicate)


def separator(predicate: Predicate = None
              ) -> Map[Iterable[Domain],
                       Tuple[Iterable[Domain], Iterable[Domain]]]:
    """
    Returns function that returns pair of iterables
    first of which consists of elements that dissatisfy given predicate
    and second one consists of elements that satisfy given predicate.

    >>> split_by_truth = separator()
    >>> tuple(map(list, split_by_truth(range(10))))
    ([0], [1, 2, 3, 4, 5, 6, 7, 8, 9])

    >>> def is_even(number: int) -> bool:
    ...     return number % 2 == 0
    >>> split_by_evenness = separator(is_even)
    >>> tuple(map(list, split_by_evenness(range(10))))
    ([1, 3, 5, 7, 9], [0, 2, 4, 6, 8])
    """
    return compose(tuple,
                   combine(scavenger(predicate), sifter(predicate)),
                   _duplicate)


def grabber(predicate: Predicate = None) -> Operator[Iterable[Domain]]:
    """
    Returns function that selects elements from the beginning of iterable
    while given predicate is satisfied.

    If predicate is not specified than true-like objects are selected.

    >>> grab_while_true_like = grabber()
    >>> list(grab_while_true_like(range(10)))
    []

    >>> from operator import gt
    >>> from functools import partial
    >>> grab_while_less_than_five = grabber(partial(gt, 5))
    >>> list(grab_while_less_than_five(range(10)))
    [0, 1, 2, 3, 4]
    """
    if predicate is None:
        predicate = bool
    return functools.partial(itertools.takewhile, predicate)


def kicker(predicate: Predicate = None) -> Operator[Iterable[Domain]]:
    """
    Returns function that skips elements from the beginning of iterable
    while given predicate is satisfied.

    If predicate is not specified than true-like objects are skipped.

    >>> kick_while_true_like = kicker()
    >>> list(kick_while_true_like(range(10)))
    [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]

    >>> from operator import gt
    >>> from functools import partial
    >>> kick_while_less_than_five = kicker(partial(gt, 5))
    >>> list(kick_while_less_than_five(range(10)))
    [5, 6, 7, 8, 9]
    """
    if predicate is None:
        predicate = bool
    return functools.partial(itertools.dropwhile, predicate)



import functools
import itertools
from collections import (abc,
                         deque)
from typing import (Iterable,
                    Tuple)

from .hints import (Domain,
                    Map)


@functools.singledispatch
def replicate(object_: Domain,
              *,
              count: int) -> Iterable[Domain]:
    """
    Returns given number of object replicas.
    """
    raise TypeError('Unsupported object type: {type}.'
                    .format(type=type(object_)))


@replicate.register(object)
# immutable strings represent a special kind of iterables
# that can be replicated by simply repeating
@replicate.register(bytes)
@replicate.register(str)
# mappings cannot be replicated as other iterables
# since they are iterable only by key
@replicate.register(abc.Mapping)
def _(object_: Domain,
      *,
      count: int) -> Iterable[Domain]:
    """
    Returns object repeated given number of times.
    """
    return itertools.repeat(object_, count)


@replicate.register(abc.Iterable)
def _replicate_iterable(object_: Iterable[Domain],
                        *,
                        count: int) -> Iterable[Iterable[Domain]]:
    """
    Returns given number of iterable replicas.
    """
    iterator = iter(object_)
    queues = [deque() for _ in itertools.repeat(None, count)]

    def replica(queue: deque) -> Iterable[Domain]:
        while True:
            if not queue:
                try:
                    element = next(iterator)
                except StopIteration:
                    return
                element_copies = replicate(element,
                                           count=count)
                for sub_queue, element_copy in zip(queues, element_copies):
                    sub_queue.append(element_copy)
            yield queue.popleft()

    yield from map(replica, queues)


@replicate.register(tuple)
def _(object_: Tuple[Domain, ...],
      *,
      count: int) -> Iterable[Tuple[Domain, ...]]:
    yield from map(tuple, _replicate_iterable(object_,
                                              count=count))


def replicator(count: int) -> Map[Domain, Iterable[Domain]]:
    """
    Returns function that replicates passed object.

    >>> triplicate = replicator(3)
    >>> list(map(tuple, triplicate(range(5))))
    [(0, 1, 2, 3, 4), (0, 1, 2, 3, 4), (0, 1, 2, 3, 4)]
    """
    return functools.partial(replicate,
                             count=count)


duplicate = replicator(2)
duplicate.__doc__ = 'Duplicates given object.'



import functools
import itertools
from operator import (attrgetter,
                      methodcaller)
from typing import (Callable,
                    Iterable,
                    List,
                    Tuple)

from paradigm import signatures
from reprit import seekers
from reprit.base import generate_repr

from . import left
from .functional import (ApplierBase,
                         compose,
                         flip)
from .hints import (Domain,
                    Map,
                    Range)
from .iterating import expand
from .reversal import reverse


def accumulator(function: Callable[[Domain, Range], Range],
                initial: Range
                ) -> Map[Iterable[Domain], Iterable[Iterable[Range]]]:
    """
    Returns function that yields cumulative results of given binary function
    starting from given initial object in direction from right to left.

    >>> def to_next_fraction(partial_denominator: int,
    ...                      reciprocal: float) -> float:
    ...     return partial_denominator + 1 / reciprocal
    >>> to_simple_continued_fractions = accumulator(to_next_fraction, 1)
    >>> from itertools import repeat
    >>> [round(fraction, 4)
    ...  for fraction in to_simple_continued_fractions(list(repeat(1, 10)))]
    [1, 2.0, 1.5, 1.6667, 1.6, 1.625, 1.6154, 1.619, 1.6176, 1.6182, 1.618]
    """
    left_accumulator = left.accumulator(flip(function), initial)
    return compose(left_accumulator, reverse)


def attacher(object_: Domain) -> Map[Iterable[Domain], Iterable[Domain]]:
    """
    Returns function that appends given object to iterable.

    >>> attach_hundred = attacher(100)
    >>> list(attach_hundred(range(10)))
    [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 100]
    """
    return functools.partial(attach,
                             object_=object_)


@functools.singledispatch
def attach(iterable: Iterable[Domain],
           object_: Domain) -> Iterable[Domain]:
    """
    Appends given object to the iterable.
    """
    yield from itertools.chain(iterable, expand(object_))


@attach.register(list)
def _(iterable: List[Domain],
      object_: Domain) -> List[Domain]:
    """
    Appends given object to the list.
    """
    return iterable + [object_]


@attach.register(tuple)
def _(iterable: Tuple[Domain, ...],
      object_: Domain) -> Tuple[Domain, ...]:
    """
    Appends given object to the tuple.
    """
    return iterable + (object_,)


def folder(function: Callable[[Domain, Range], Range],
           initial: Range) -> Map[Iterable[Domain], Range]:
    """
    Returns function that cumulatively applies given binary function
    starting from given initial object in direction from right to left.

    >>> to_sum_evaluation_order = folder('({} + {})'.format, 0)
    >>> to_sum_evaluation_order(range(1, 10))
    '(1 + (2 + (3 + (4 + (5 + (6 + (7 + (8 + (9 + 0)))))))))'
    """
    left_folder = left.folder(flip(function), initial)
    return compose(left_folder, reverse)


class Applier(ApplierBase):
    def __init__(self, function: Callable[..., Range],
                 *args: Domain,
                 **kwargs: Domain) -> None:
        super().__init__(function, *args, **kwargs)

    def __call__(self, *args: Domain, **kwargs: Domain) -> Range:
        return self.func(*args, *self.args, **self.keywords, **kwargs)

    __repr__ = generate_repr(__init__,
                             field_seeker=seekers.complex_)


@signatures.factory.register(Applier)
def _(object_: Applier) -> signatures.Base:
    return _bind_positionals(signatures.factory(object_.func)
                             .bind(**object_.keywords),
                             object_.args)


@functools.singledispatch
def _bind_positionals(signature: signatures.Base,
                      args: Tuple[Domain, ...]) -> signatures.Base:
    raise TypeError('Unsupported signature type: {type}.'
                    .format(type=type(signature)))


@_bind_positionals.register(signatures.Plain)
def _(signature: signatures.Plain,
      args: Tuple[Domain, ...]) -> signatures.Base:
    if not args:
        return signature
    variadic_positionals = signature.parameters_by_kind[
        signatures.Parameter.Kind.VARIADIC_POSITIONAL]
    positionals = (signature.parameters_by_kind[
                       signatures.Parameter.Kind.POSITIONAL_ONLY]
                   + signature.parameters_by_kind[
                       signatures.Parameter.Kind.POSITIONAL_OR_KEYWORD])
    if len(args) > len(positionals) and not variadic_positionals:
        value = 'argument' + 's' * (len(positionals) != 1)
        raise TypeError('Takes {parameters_count} positional {value}, '
                        'but {arguments_count} {verb} given.'
                        .format(parameters_count=len(positionals),
                                value=value,
                                arguments_count=len(args),
                                verb='was' if len(args) == 1 else 'were'))
    non_positionals = (signature.parameters_by_kind[
                           signatures.Parameter.Kind.KEYWORD_ONLY]
                       + signature.parameters_by_kind[
                           signatures.Parameter.Kind.VARIADIC_KEYWORD])
    signatures_parameters = []
    if len(args) <= len(positionals):
        signatures_parameters.append(positionals[:-len(args)]
                                     + non_positionals)
    if variadic_positionals:
        signatures_parameters.append(signature.parameters)
        for limit in range(1, len(args)):
            signatures_parameters.append(positionals[:-limit]
                                         + variadic_positionals
                                         + non_positionals)
    positionals_or_keywords = signature.parameters_by_kind[
        signatures.Parameter.Kind.POSITIONAL_OR_KEYWORD]
    positionals_or_keywords_with_defaults_count = sum(
            map(attrgetter('has_default'), positionals_or_keywords))
    for offset in range(1, positionals_or_keywords_with_defaults_count + 1):
        signatures_parameters.append(
                positionals[:-(len(args) + offset)]
                + [signatures.Parameter(
                        name=parameter.name,
                        kind=signatures.Parameter.Kind.KEYWORD_ONLY,
                        has_default=parameter.has_default)
                    for parameter in positionals_or_keywords[-offset:]]
                + non_positionals)
    return signatures.Overloaded(*(signatures.Plain(*parameters)
                                   for parameters in signatures_parameters))


@_bind_positionals.register(signatures.Overloaded)
def _(signature: signatures.Overloaded,
      args: Tuple[Domain, ...]) -> signatures.Base:
    sub_signatures = list(filter(methodcaller(signatures.Base.expects.__name__,
                                              *args),
                                 signature.signatures))
    if not sub_signatures:
        raise TypeError('No corresponding signature found.')
    return signatures.Overloaded(*map(functools.partial(_bind_positionals,
                                                        args=args),
                                      sub_signatures))


def applier(function: Callable[..., Range],
            *args: Domain,
            **kwargs: Domain) -> Callable[..., Range]:
    """
    Returns function that behaves like given function
    with given arguments partially applied.
    Given positional arguments will be added to the right end.

    >>> square = applier(pow, 2)
    >>> square(10)
    100
    """
    return Applier(function, *args, **kwargs)



import functools
from collections import ChainMap
from typing import (Callable,
                    Iterable,
                    Optional,
                    Union)

from .hints import (Domain,
                    Map,
                    Operator,
                    Sortable)

Key = Map[Domain, Sortable]
Implementation = Callable[..., Iterable[Domain]]

stable_implementations = {}
unstable_implementations = {}
implementations = ChainMap(stable_implementations,
                           unstable_implementations)


def search_implementation(algorithm: str) -> Implementation:
    try:
        return implementations[algorithm]
    except KeyError as error:
        algorithms_str = ', '.join(map(repr, implementations))
        error_message = ('Algorithm is not found: {algorithm}. '
                         'Available algorithms are: {algorithms}.'
                         .format(algorithm=algorithm,
                                 algorithms=algorithms_str))
        raise LookupError(error_message) from error


def register_implementation(algorithm: str,
                            implementation: Optional[Implementation] = None,
                            *,
                            stable: bool = False,
                            overwrite: bool = False
                            ) -> Union[Operator[Implementation],
                                       Implementation]:
    """
    Registers implementation of sorting algorithm.

    >>> from typing import Any
    >>> @register_implementation('CUSTOMSORTING')
    ... def custom_sorting(iterable: Iterable[Domain],
    ...                    *,
    ...                    key: Optional[Map[Domain, Any]] = None
    ...                    ) -> Iterable[Domain]:
    ...     ...
    """
    if implementation is None:
        return functools.partial(register_implementation, algorithm,
                                 stable=stable)
    if algorithm in implementations:
        if overwrite:
            del implementations[algorithm]
        else:
            raise ValueError('Algorithm "{algorithm}" '
                             'has been already registered.'
                             .format(algorithm=algorithm))
    if stable:
        stable_implementations[algorithm] = implementation
    else:
        unstable_implementations[algorithm] = implementation
    return implementation


DEFAULT_ALGORITHM = 'TIMSORT'
register_implementation(DEFAULT_ALGORITHM, sorted,
                        stable=True)


def sorter(*,
           algorithm: str = DEFAULT_ALGORITHM,
           key: Optional[Key] = None) -> Operator[Iterable[Domain]]:
    """
    Returns function that generates sorted iterable
    by given key with specified algorithm.

    >>> sort = sorter()
    >>> sort('Hello World!')
    [' ', '!', 'H', 'W', 'd', 'e', 'l', 'l', 'l', 'o', 'o', 'r']
    """
    implementation = search_implementation(algorithm)
    return functools.partial(implementation,
                             key=key)



from collections import abc
from typing import (Callable,
                    Container,
                    Iterable,
                    TypeVar)

try:
    from typing import Protocol
except ImportError:
    from typing_extensions import Protocol

Domain = TypeVar('Domain')
Range = TypeVar('Range')
Map = Callable[[Domain], Range]
Operator = Map[Domain, Domain]
Predicate = Map[Domain, bool]


class Sortable(Protocol):
    def __lt__(self, other: 'Sortable') -> bool:
        pass


try:
    from collections.abc import Collection
    from typing import Collection as FiniteIterable
except ImportError:
    # for Python 3.5
    class Collection(abc.Sized, abc.Iterable, abc.Container):
        __slots__ = ()

        @classmethod
        def __subclasshook__(cls, subclass: type) -> bool:
            if cls is Collection:
                mro = subclass.__mro__
                for method in ('__len__', '__iter__', '__contains__'):
                    for base_class in mro:
                        if method in base_class.__dict__:
                            if base_class.__dict__[method] is None:
                                return NotImplemented
                            break
                    else:
                        return NotImplemented
                return True
            return NotImplemented


    class FiniteIterable(abc.Sized, Iterable[Domain], Container[Domain],
                         extra=Collection):
        __slots__ = ()



import functools
from collections import (abc,
                         deque)
from typing import (Iterable,
                    overload)

from .hints import (Collection,
                    Domain,
                    FiniteIterable,
                    Range)


@overload
def transpose(object_: Iterable[FiniteIterable[Domain]]
              ) -> FiniteIterable[Iterable[Domain]]:
    pass


@overload
def transpose(object_: FiniteIterable[Iterable[Domain]]
              ) -> Iterable[FiniteIterable[Domain]]:
    pass


@functools.singledispatch
def transpose(object_: Domain) -> Range:
    """
    Transposes given object.

    >>> list(map(tuple, transpose(zip(range(10), range(10, 20)))))
    [(0, 1, 2, 3, 4, 5, 6, 7, 8, 9), (10, 11, 12, 13, 14, 15, 16, 17, 18, 19)]
    """
    raise TypeError('Unsupported object type: {type}.'
                    .format(type=type))


@transpose.register(abc.Iterable)
def _(object_: Iterable[FiniteIterable[Domain]]
      ) -> FiniteIterable[Iterable[Domain]]:
    """
    Transposes given iterable of finite iterables.
    """
    iterator = iter(object_)
    try:
        first_elements = next(iterator)
    except StopIteration:
        return ()
    queues = [deque([element])
              for element in first_elements]

    def coordinate(queue: deque) -> Iterable[Domain]:
        while True:
            if not queue:
                try:
                    elements = next(iterator)
                except StopIteration:
                    return
                for sub_queue, element in zip(queues, elements):
                    sub_queue.append(element)
            yield queue.popleft()

    return tuple(map(coordinate, queues))


@transpose.register(Collection)
def _(object_: FiniteIterable[Iterable[Domain]]
      ) -> Iterable[FiniteIterable[Domain]]:
    """
    Transposes given finite iterable of iterables.
    """
    yield from zip(*object_)



from operator import (not_,
                      xor)

from . import left
from .functional import (cleave,
                         compose)
from .hints import Predicate


def conjoin(*predicates: Predicate) -> Predicate:
    """
    Returns conjunction of given predicates.

    >>> is_valid_constant_identifier = conjoin(str.isupper, str.isidentifier)
    >>> is_valid_constant_identifier('SECOND_SECTION')
    True
    >>> is_valid_constant_identifier('2ND_SECTION')
    False
    """
    return compose(all, cleave(*predicates))


def disjoin(*predicates: Predicate) -> Predicate:
    """
    Returns disjunction of given predicates.

    >>> alphabetic_or_numeric = disjoin(str.isalpha, str.isnumeric)
    >>> alphabetic_or_numeric('Hello')
    True
    >>> alphabetic_or_numeric('42')
    True
    >>> alphabetic_or_numeric('Hello42')
    False
    """
    return compose(any, cleave(*predicates))


def exclusive_disjoin(*predicates: Predicate) -> Predicate:
    """
    Returns exclusive disjunction of given predicates.

    >>> from keyword import iskeyword
    >>> valid_object_name = exclusive_disjoin(str.isidentifier, iskeyword)
    >>> valid_object_name('valid_object_name')
    True
    >>> valid_object_name('_')
    True
    >>> valid_object_name('1')
    False
    >>> valid_object_name('lambda')
    False
    """
    return compose(left.folder(xor, False), cleave(*predicates))


def negate(predicate: Predicate) -> Predicate:
    """
    Returns negated version of given predicate.

    >>> from lz.logical import negate
    >>> false_like = negate(bool)
    >>> false_like([])
    True
    >>> false_like([0])
    False
    """
    return compose(not_, predicate)



from numbers import Real
from typing import overload


@overload
def ceil_division(dividend: int, divisor: int) -> int:
    pass


def ceil_division(dividend: Real, divisor: Real) -> Real:
    """
    Divides given numbers with ceiling.

    >>> ceil_division(10, 2)
    5
    >>> ceil_division(10, -2)
    -5
    >>> ceil_division(10, 3)
    4
    >>> ceil_division(10, -3)
    -3
    >>> ceil_division(-10, -3)
    4
    >>> ceil_division(-10, 3)
    -3
    """
    return -(-dividend // divisor)



from functools import partial
from operator import attrgetter
from typing import (Any,
                    Tuple)

from .hints import Predicate


def instance_of(*types: type) -> Predicate:
    """
    Creates predicate that checks if object is instance of given types.

    >>> is_any_string = instance_of(str, bytes, bytearray)
    >>> is_any_string(b'')
    True
    >>> is_any_string('')
    True
    >>> is_any_string(1)
    False
    """
    predicate = partial(is_instance_of,
                        types=types)
    types_names = list(map(attrgetter('__name__'), types))
    predicate.__name__ = 'is_instance_of_' + '_or_'.join(types_names)
    predicate.__qualname__ = 'is_instance_of_' + '_or_'.join(types_names)
    types_full_names = map(attrgetter('__qualname__'), types)
    predicate.__doc__ = ('Checks if given object is instance '
                         'of one of types: "{types}".'
                         .format(types='", "'.join(types_full_names)))
    return predicate


def subclass_of(*types: type) -> Predicate:
    """
    Creates predicate that checks if type is subclass of given types.

    >>> is_metaclass = subclass_of(type)
    >>> is_metaclass(type)
    True
    >>> is_metaclass(object)
    False
    """
    predicate = partial(is_subclass_of,
                        types=types)
    types_names = list(map(attrgetter('__name__'), types))
    predicate.__name__ = 'is_subclass_of_' + '_or_'.join(types_names)
    predicate.__qualname__ = 'is_subclass_of_' + '_or_'.join(types_names)
    types_full_names = map(attrgetter('__qualname__'), types)
    predicate.__doc__ = ('Checks if given type is subclass '
                         'of one of types: "{types}".'
                         .format(types='", "'.join(types_full_names)))
    return predicate


def is_instance_of(object_: Any, types: Tuple[type, ...]) -> bool:
    return isinstance(object_, types)


def is_subclass_of(type_: type, types: Tuple[type, ...]) -> bool:
    return issubclass(type_, types)



import ast
import functools
import inspect
import itertools
import sys
from collections import abc
from types import (MappingProxyType,
                   MethodType)
from typing import (Any,
                    Callable,
                    Dict,
                    Iterable,
                    Optional,
                    Tuple,
                    Type,
                    Union)

from paradigm import signatures
from reprit import seekers
from reprit.base import generate_repr

from .hints import (Domain,
                    Map,
                    Range)


def identity(argument: Domain) -> Domain:
    """
    Returns object itself.

    >>> identity(0)
    0
    """
    return argument


def compose(last_function: Map[Any, Range],
            *front_functions: Callable[..., Any]) -> Callable[..., Range]:
    """
    Returns functions composition.

    >>> sum_of_first_n_natural_numbers = compose(sum, range)
    >>> sum_of_first_n_natural_numbers(10)
    45
    """
    caller_frame_info = inspect.stack()[1]
    return Composition(last_function, *front_functions,
                       file_path=caller_frame_info.filename,
                       line_number=caller_frame_info.lineno,
                       line_offset=0)


class Composition:
    def __new__(cls,
                *functions: Callable[..., Any],
                **kwargs: Any) -> Union['Composition', Callable[..., Range]]:
        if len(functions) == 1:
            return functions[0]
        return super().__new__(cls)

    def __init__(self,
                 *functions: Callable[..., Any],
                 file_path: Optional[str] = None,
                 line_number: int = 0,
                 line_offset: int = 0) -> None:
        def flatten(function: Callable[..., Any]
                    ) -> Iterable[Callable[..., Any]]:
            if isinstance(function, type(self)):
                yield from function.functions
            else:
                yield function

        self._functions = tuple(itertools.chain
                                .from_iterable(map(flatten, functions)))
        self._function = None
        if file_path is None:
            file_path = __file__
        self._file_path = file_path
        self._line_number = line_number
        self._line_offset = line_offset

    @property
    def functions(self) -> Tuple[Callable[..., Any], ...]:
        return self._functions

    @property
    def function(self) -> Callable[..., Range]:
        if self._function is None:
            self._function = _compose(*self.functions,
                                      function_name='composition',
                                      file_path=self._file_path,
                                      line_number=self._line_number,
                                      line_offset=self._line_offset)
        return self._function

    def __call__(self, *args: Domain, **kwargs: Domain) -> Range:
        return self.function(*args, **kwargs)

    def __get__(self,
                instance: Domain,
                owner: Type[Domain]) -> Callable[..., Range]:
        return MethodType(self.function, instance)

    __repr__ = generate_repr(__init__,
                             field_seeker=seekers.complex_)


@signatures.factory.register(Composition)
def _(object_: Composition) -> signatures.Base:
    return signatures.factory(object_.functions[-1])


def _compose(*functions: Callable[..., Any],
             function_name: str,
             arguments_factory: Callable[..., ast.arguments] =
             ast.arguments if sys.version_info < (3, 8)
             # Python3.8 adds positional-only arguments
             else functools.partial(ast.arguments, []),
             module_factory: Callable[..., ast.Module] =
             ast.Module if sys.version_info < (3, 8)
             # Python3.8 adds `type_ignores` parameter
             else functools.partial(ast.Module,
                                    type_ignores=[]),
             file_path: str,
             line_number: int,
             line_offset: int) -> Callable[..., Range]:
    def function_to_unique_name(function: Callable) -> str:
        # we are not using `__name__`/`__qualname__` attributes
        # due to their potential non-uniqueness
        return '_' + str(id(function)).replace('-', '_')

    functions_names = list(map(function_to_unique_name, functions))

    set_attributes = functools.partial(functools.partial,
                                       lineno=line_number,
                                       col_offset=line_offset)

    variadic_positionals_name = 'args'
    variadic_keywords_name = 'kwargs'

    def to_next_call_node(node: ast.Call, name: str) -> ast.Call:
        return set_attributes(ast.Call)(to_name_node(name), [node], [])

    def to_name_node(name: str,
                     *,
                     context_factory: Type[ast.expr_context] = ast.Load
                     ) -> ast.Name:
        return set_attributes(ast.Name)(name, context_factory())

    reversed_functions_names = reversed(functions_names)
    calls_node = set_attributes(ast.Call)(
            to_name_node(next(reversed_functions_names)),
            [set_attributes(ast.Starred)(
                    to_name_node(variadic_positionals_name),
                    ast.Load())],
            [ast.keyword(None,
                         to_name_node(variadic_keywords_name))])
    calls_node = functools.reduce(to_next_call_node,
                                  reversed_functions_names,
                                  calls_node)
    function_definition_node = set_attributes(ast.FunctionDef)(
            function_name,
            arguments_factory(
                    [],
                    set_attributes(ast.arg)(variadic_positionals_name, None),
                    [],
                    [],
                    set_attributes(ast.arg)(variadic_keywords_name, None),
                    []),
            [set_attributes(ast.Return)(calls_node)],
            [],
            None)
    tree = module_factory([function_definition_node])
    code = compile(tree, file_path, 'exec')
    namespace = dict(zip(functions_names, functions))
    exec(code, namespace)
    return namespace[function_name]


def combine(*maps: Map) -> Map[Iterable[Domain], Iterable[Range]]:
    """
    Returns function that applies each map to corresponding argument.

    >>> encoder_decoder = combine(str.encode, bytes.decode)
    >>> list(encoder_decoder(['hello', b'world']))
    [b'hello', 'world']
    """
    return Combination(*maps)


class Combination:
    def __init__(self, *maps: Map) -> None:
        self.maps = maps

    def __call__(self, arguments: Iterable[Domain]) -> Iterable[Range]:
        yield from (map_(argument)
                    for map_, argument in zip(self.maps, arguments))

    __repr__ = generate_repr(__init__)


@signatures.factory.register(Combination)
def _(object_: Combination) -> signatures.Base:
    return signatures.factory(object_.__call__)


class ApplierBase(abc.Callable):
    def __init__(self, function: Callable[..., Range],
                 *args: Domain,
                 **kwargs: Domain) -> None:
        if isinstance(function, type(self)):
            args = function.args + args
            kwargs = {**function.keywords, **kwargs}
            function = function.func
        self._function = function
        self._args = args
        self._kwargs = kwargs

    @property
    def func(self) -> Callable[..., Range]:
        return self._function

    @property
    def args(self) -> Tuple[Domain, ...]:
        return self._args

    @property
    def keywords(self) -> Dict[str, Domain]:
        return self._kwargs


ApplierBase.register(functools.partial)


class Curry(ApplierBase):
    def __init__(self,
                 function: Callable[..., Range],
                 signature: signatures.Base,
                 *args: Domain,
                 **kwargs: Domain) -> None:
        super().__init__(function, *args, **kwargs)
        self.signature = signature

    def __call__(self, *args: Domain, **kwargs: Domain
                 ) -> Union['Curry', Range]:
        args = self.args + args
        kwargs = {**self.keywords, **kwargs}
        try:
            return self.func(*args, **kwargs)
        except TypeError:
            if (not self.signature.expects(*args, **kwargs)
                    or self.signature.all_set(*args, **kwargs)):
                raise
        return type(self)(self.func, self.signature, *args, **kwargs)

    __repr__ = generate_repr(__init__,
                             field_seeker=seekers.complex_)


@signatures.factory.register(Curry)
def _(object_: Curry) -> signatures.Base:
    return object_.signature


def curry(function: Callable[..., Range],
          *,
          signature: Optional[signatures.Base] = None) -> Curry:
    """
    Returns curried version of given function.

    >>> curried_pow = curry(pow)
    >>> two_to_power = curried_pow(2)
    >>> two_to_power(10)
    1024
    """
    if signature is None:
        signature = signatures.factory(function)
    return Curry(function, signature)


def pack(function: Callable[..., Range]) -> Map[Iterable[Domain], Range]:
    """
    Returns function that works with single iterable parameter
    by unpacking elements to given function.

    >>> packed_int = pack(int)
    >>> packed_int(['10'])
    10
    >>> packed_int(['10'], {'base': 2})
    2
    """
    return functools.partial(apply, function)


def apply(function: Callable[..., Range],
          args: Iterable[Domain],
          kwargs: Dict[str, Any] = MappingProxyType({})) -> Range:
    """
    Calls given function with given positional and keyword arguments.
    """
    return function(*args, **kwargs)


def to_constant(object_: Domain) -> Callable[..., Domain]:
    """
    Returns function that always returns given object.

    >>> always_zero = to_constant(0)
    >>> always_zero()
    0
    >>> always_zero(1)
    0
    >>> always_zero(how_about=2)
    0
    """
    return Constant(object_)


class Constant:
    def __init__(self, object_: Domain) -> None:
        self.object_ = object_

    def __call__(self, *args: Any, **kwargs: Any) -> Domain:
        return self.object_

    __repr__ = generate_repr(__init__)


@signatures.factory.register(Constant)
def _(object_: Constant) -> signatures.Base:
    return signatures.factory(object_.__call__)


def flip(function: Callable[..., Range]) -> Callable[..., Range]:
    """
    Returns function with positional arguments flipped.

    >>> flipped_power = flip(pow)
    >>> flipped_power(2, 4)
    16
    """
    return functools.partial(call_flipped, function)


def call_flipped(function: Callable[..., Range],
                 *args: Domain,
                 **kwargs: Domain) -> Range:
    """
    Calls given function with positional arguments flipped.
    """
    return function(*args[::-1], **kwargs)


def cleave(*functions: Callable[..., Range]) -> Callable[..., Iterable[Range]]:
    """
    Returns function that separately applies
    given functions to the same arguments.

    >>> to_min_and_max = cleave(min, max)
    >>> list(to_min_and_max(range(10)))
    [0, 9]
    >>> list(to_min_and_max(range(0), default=None))
    [None, None]
    """
    return Cleavage(*functions)


class Cleavage:
    def __init__(self, *functions: Map) -> None:
        self.functions = functions

    def __call__(self, *args: Domain, **kwargs: Domain) -> Iterable[Range]:
        yield from (function(*args, **kwargs)
                    for function in self.functions)

    __repr__ = generate_repr(__init__)


@signatures.factory.register(Cleavage)
def _(object_: Cleavage) -> signatures.Base:
    return signatures.factory(object_.__call__)


def flatmap(function: Map[Domain, Iterable[Range]],
            *iterables: Iterable[Domain]) -> Iterable[Range]:
    """
    Applies given function to the arguments aggregated from given iterables
    and concatenates results into plain iterable.

    >>> list(flatmap(range, range(5)))
    [0, 0, 1, 0, 1, 2, 0, 1, 2, 3]
    """
    yield from itertools.chain.from_iterable(map(function, *iterables))



import functools
import itertools
from typing import (Callable,
                    Iterable,
                    List,
                    Tuple)

from .hints import (Domain,
                    Map,
                    Range)
from .iterating import expand


def accumulator(function: Callable[[Range, Domain], Range],
                initial: Range
                ) -> Map[Iterable[Domain], Iterable[Range]]:
    """
    Returns function that yields cumulative results of given binary function
    starting from given initial object in direction from left to right.

    >>> import math
    >>> to_pi_approximations = accumulator(round, math.pi)
    >>> list(to_pi_approximations(range(5, 0, -1)))
    [3.141592653589793, 3.14159, 3.1416, 3.142, 3.14, 3.1]
    """
    return functools.partial(accumulate,
                             function=function,
                             initial=initial)


def accumulate(iterable: Iterable[Domain],
               function: Callable[[Range, Domain], Range],
               initial: Range) -> Iterable[Range]:
    """
    Yields cumulative results of given binary function
    starting from given initial object in direction from left to right.
    """
    yield from itertools.accumulate(attach(iterable, initial), function)


def attacher(object_: Domain) -> Map[Iterable[Domain], Iterable[Domain]]:
    """
    Returns function that prepends given object to iterable.

    >>> attach_hundred = attacher(100)
    >>> list(attach_hundred(range(10)))
    [100, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
    """
    return functools.partial(attach,
                             object_=object_)


@functools.singledispatch
def attach(iterable: Iterable[Domain],
           object_: Domain) -> Iterable[Domain]:
    """
    Prepends given object to the iterable.
    """
    yield from itertools.chain(expand(object_), iterable)


@attach.register(list)
def _(iterable: List[Domain],
      object_: Domain) -> List[Domain]:
    """
    Prepends given object to the list.
    """
    return [object_] + iterable


@attach.register(tuple)
def _(iterable: Tuple[Domain, ...],
      object_: Domain) -> Tuple[Domain, ...]:
    """
    Prepends given object to the tuple.
    """
    return (object_,) + iterable


def folder(function: Callable[[Range, Domain], Range],
           initial: Range) -> Map[Iterable[Domain], Range]:
    """
    Returns function that cumulatively applies given binary function
    starting from given initial object in direction from left to right.

    >>> to_sum_evaluation_order = folder('({} + {})'.format, 0)
    >>> to_sum_evaluation_order(range(1, 10))
    '(((((((((0 + 1) + 2) + 3) + 4) + 5) + 6) + 7) + 8) + 9)'
    """
    return functools.partial(fold,
                             function=function,
                             initial=initial)


def fold(iterable: Iterable[Domain],
         function: Callable[[Range, Domain], Range],
         initial: Range) -> Range:
    """
    Cumulatively applies given binary function
    starting from given initial object in direction from left to right.
    """
    return functools.reduce(function, iterable, initial)


Applier = functools.partial


def applier(function: Callable[..., Range],
            *args: Domain,
            **kwargs: Domain) -> Callable[..., Range]:
    """
    Returns function that behaves like given function
    with given arguments partially applied.
    Given positional arguments will be added to the left end.

    >>> count_from_zero_to = applier(range, 0)
    >>> list(count_from_zero_to(10))
    [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
    """
    return Applier(function, *args, **kwargs)



import functools
import itertools
from collections import (OrderedDict,
                         abc,
                         deque)
from operator import is_not
from typing import (Any,
                    Hashable,
                    Iterable,
                    MutableMapping,
                    Sequence,
                    Sized,
                    Tuple,
                    Type)

from .functional import flatmap
from .hints import (Domain,
                    Map,
                    Operator,
                    Range)


@functools.singledispatch
def capacity(iterable: Iterable[Any]) -> int:
    """
    Returns number of elements in iterable.

    >>> capacity(range(0))
    0
    >>> capacity(range(10))
    10
    """
    counter = itertools.count()
    # order matters: if `counter` goes first,
    # then it will be incremented even for empty `iterable`
    deque(zip(iterable, counter),
          maxlen=0)
    return next(counter)


@capacity.register(abc.Sized)
def _(iterable: Sized) -> int:
    """
    Returns number of elements in sized iterable.
    """
    return len(iterable)


def first(iterable: Iterable[Domain]) -> Domain:
    """
    Returns first element of iterable.

    >>> first(range(10))
    0
    """
    try:
        return next(iter(iterable))
    except StopIteration as error:
        raise ValueError('Argument supposed to be non-empty.') from error


def last(iterable: Iterable[Domain]) -> Domain:
    """
    Returns last element of iterable.

    >>> last(range(10))
    9
    """
    try:
        return deque(iterable,
                     maxlen=1)[0]
    except IndexError as error:
        raise ValueError('Argument supposed to be non-empty.') from error


def cut(iterable: Iterable[Domain],
        *,
        slice_: slice) -> Iterable[Domain]:
    """
    Selects elements from iterable based on given slice.

    Slice fields supposed to be unset or non-negative
    since it is hard to evaluate negative indices/step for arbitrary iterable
    which may be potentially infinite
    or change previous elements if iterating made backwards.
    """
    yield from itertools.islice(iterable,
                                slice_.start, slice_.stop, slice_.step)


def cutter(slice_: slice) -> Operator[Iterable[Domain]]:
    """
    Returns function that selects elements from iterable based on given slice.

    >>> to_first_triplet = cutter(slice(3))
    >>> list(to_first_triplet(range(10)))
    [0, 1, 2]

    >>> to_second_triplet = cutter(slice(3, 6))
    >>> list(to_second_triplet(range(10)))
    [3, 4, 5]

    >>> cut_out_every_third = cutter(slice(0, None, 3))
    >>> list(cut_out_every_third(range(10)))
    [0, 3, 6, 9]
    """
    result = functools.partial(cut,
                               slice_=slice_)
    result.__doc__ = ('Selects elements from iterable {slice}.'
                      .format(slice=_slice_to_description(slice_)))
    return result


def _slice_to_description(slice_: slice) -> str:
    """Generates human readable representation of `slice` object."""
    slice_description_parts = []
    start_is_specified = bool(slice_.start)
    if start_is_specified:
        slice_description_parts.append('starting from position {start}'
                                       .format(start=slice_.start))
    step_is_specified = slice_.step is not None
    if step_is_specified:
        slice_description_parts.append('with step {step}'
                                       .format(step=slice_.step))
    if slice_.stop is not None:
        stop_description_part = ('stopping at position {stop}'
                                 .format(stop=slice_.stop))
        if start_is_specified or step_is_specified:
            stop_description_part = 'and ' + stop_description_part
        slice_description_parts.append(stop_description_part)
    return ' '.join(slice_description_parts)


def chopper(size: int) -> Map[Iterable[Domain], Iterable[Sequence[Domain]]]:
    """
    Returns function that splits iterable into chunks of given size.

    >>> in_three = chopper(3)
    >>> list(map(tuple, in_three(range(10))))
    [(0, 1, 2), (3, 4, 5), (6, 7, 8), (9,)]
    """
    result = functools.partial(chop,
                               size=size)
    result.__doc__ = ('Splits iterable into chunks of size {size}.\n'
                      .format(size=size))
    return result


@functools.singledispatch
def chop(iterable: Iterable[Domain],
         *,
         size: int) -> Iterable[Sequence[Domain]]:
    """
    Splits iterable into chunks of given size.
    """
    iterator = iter(iterable)
    yield from iter(lambda: tuple(itertools.islice(iterator, size)), ())


@chop.register(abc.Sequence)
def _(iterable: Sequence[Domain],
      *,
      size: int) -> Iterable[Sequence[Domain]]:
    """
    Splits sequence into chunks of given size.
    """
    if not size:
        return
    for start in range(0, len(iterable), size):
        yield iterable[start:start + size]


# deque do not support slice notation
chop.register(deque, chop.registry[object])

in_two = chopper(2)
in_three = chopper(3)
in_four = chopper(4)


def slide(iterable: Iterable[Domain],
          *,
          size: int) -> Iterable[Tuple[Domain, ...]]:
    """
    Slides over iterable with window of given size.
    """
    iterator = iter(iterable)
    initial = tuple(itertools.islice(iterator, size))

    def shift(previous: Tuple[Domain, ...],
              element: Domain) -> Tuple[Domain, ...]:
        return previous[1:] + (element,)

    yield from itertools.accumulate(itertools.chain([initial], iterator),
                                    shift)


def slider(size: int) -> Map[Iterable[Domain], Iterable[Tuple[Domain, ...]]]:
    """
    Returns function that slides over iterable with window of given size.

    >>> pairwise = slider(2)
    >>> list(pairwise(range(10)))
    [(0, 1), (1, 2), (2, 3), (3, 4), (4, 5), (5, 6), (6, 7), (7, 8), (8, 9)]
    """
    result = functools.partial(slide,
                               size=size)
    result.__doc__ = ('Slides over iterable with window of size {size}.'
                      .format(size=size))
    return result


pairwise = slider(2)
triplewise = slider(3)
quadruplewise = slider(4)


def header(size: int) -> Operator[Iterable[Domain]]:
    """
    Returns function that selects elements from the beginning of iterable.
    Resulted iterable will have size not greater than given one.

    >>> to_first_pair = header(2)
    >>> list(to_first_pair(range(10)))
    [0, 1]
    """
    result = cutter(slice(size))
    result.__doc__ = ('Selects {size} elements from the beginning of iterable.'
                      .format(size=size))
    return result


@functools.singledispatch
def trail(iterable: Iterable[Domain],
          *,
          size: int) -> Iterable[Domain]:
    """
    Selects elements from the end of iterable.
    Resulted iterable will have size not greater than given one.
    """
    return deque(iterable,
                 maxlen=size)


@trail.register(abc.Sequence)
def _(iterable: Sequence[Domain],
      *,
      size: int) -> Sequence[Domain]:
    """
    Selects elements from the end of sequence.
    Resulted sequence will have size not greater than given one.
    """
    return iterable[-size:] if size else iterable[:size]


# deque do not support slice notation
trail.register(deque, trail.registry[object])


def trailer(size: int) -> Operator[Iterable[Domain]]:
    """
    Returns function that selects elements from the end of iterable.
    Resulted iterable will have size not greater than given one.

    >>> to_last_pair = trailer(2)
    >>> list(to_last_pair(range(10)))
    [8, 9]
    """
    result = functools.partial(trail,
                               size=size)
    result.__doc__ = ('Selects {size} elements from the end of iterable.'
                      .format(size=size))
    return result


def mapper(map_: Map) -> Map[Iterable[Domain], Iterable[Range]]:
    """
    Returns function that applies given map to the each element of iterable.

    >>> to_str = mapper(str)
    >>> list(to_str(range(10)))
    ['0', '1', '2', '3', '4', '5', '6', '7', '8', '9']
    """
    return functools.partial(map, map_)


def flatmapper(map_: Map[Domain, Iterable[Range]]
               ) -> Map[Iterable[Domain], Iterable[Range]]:
    """
    Returns function that applies map to the each element of iterable
    and flattens results.

    >>> relay = flatmapper(range)
    >>> list(relay(range(5)))
    [0, 0, 1, 0, 1, 2, 0, 1, 2, 3]
    """
    return functools.partial(flatmap, map_)


Group = Tuple[Hashable, Iterable[Domain]]


def group_by(iterable: Iterable[Domain],
             *,
             key: Map[Domain, Hashable],
             mapping_cls: Type[MutableMapping]) -> Iterable[Group]:
    """
    Groups iterable elements based on given key.
    """
    groups = mapping_cls()
    for element in iterable:
        groups.setdefault(key(element), []).append(element)
    yield from groups.items()


def grouper(key: Map[Domain, Hashable],
            *,
            mapping_cls: Type[MutableMapping] = OrderedDict
            ) -> Map[Iterable[Domain], Iterable[Group]]:
    """
    Returns function that groups iterable elements based on given key.

    >>> group_by_absolute_value = grouper(abs)
    >>> list(group_by_absolute_value(range(-5, 5)))
    [(5, [-5]), (4, [-4, 4]), (3, [-3, 3]), (2, [-2, 2]), (1, [-1, 1]), (0, [0])]

    >>> def modulo_two(number: int) -> int:
    ...     return number % 2
    >>> group_by_evenness = grouper(modulo_two)
    >>> list(group_by_evenness(range(10)))
    [(0, [0, 2, 4, 6, 8]), (1, [1, 3, 5, 7, 9])]
    """
    return functools.partial(group_by,
                             key=key,
                             mapping_cls=mapping_cls)


def expand(object_: Domain) -> Iterable[Domain]:
    """
    Wraps object into iterable.

    >>> list(expand(0))
    [0]
    """
    yield object_


def flatten(iterable: Iterable[Iterable[Domain]]) -> Iterable[Domain]:
    """
    Returns plain iterable from iterable of iterables.

    >>> list(flatten([range(5), range(10, 20)]))
    [0, 1, 2, 3, 4, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19]
    """
    yield from itertools.chain.from_iterable(iterable)


def interleave(iterable: Iterable[Iterable[Domain]]) -> Iterable[Domain]:
    """
    Interleaves elements from given iterable of iterables.

    >>> list(interleave([range(5), range(10, 20)]))
    [0, 10, 1, 11, 2, 12, 3, 13, 4, 14, 15, 16, 17, 18, 19]
    """
    iterators = itertools.cycle(map(iter, iterable))
    while True:
        try:
            for iterator in iterators:
                yield next(iterator)
        except StopIteration:
            is_not_exhausted = functools.partial(is_not, iterator)
            iterators = itertools.cycle(itertools.takewhile(is_not_exhausted,
                                                            iterators))
        else:
            return
